/* Copyright (c) 1993 by The Johns Hopkins University */


/* METRIC.C:  Routines for defining the distance metric	*/
/*   		between two instances			*/





#include <stdio.h>
#include <math.h>

#include "config.h"
#include "pebls.h"

extern config_type CONFIG;				    
extern int count[CLASSES_MAX+1][FEATURES_MAX][VALUES_MAX];  
extern float dtable[FEATURES_MAX][VALUES_MAX][VALUES_MAX];
extern float cprob[CLASSES_MAX][FEATURES_MAX][VALUES_MAX];
extern instance_type data[INSTANCES_MAX];
extern int class_count[CLASSES_MAX];
extern float prior[CLASSES_MAX];


/* ------------------------------------------------------------ */
/* INITIALIZE_TRAINING: Clear Value Counts Statistics		*/

void initialize_training(void)
{
    int c,f,v,i;
    int instances = CONFIG.instances;

    for (c=0; c<CLASSES_MAX+1; c++)
      {
	class_count[c] = 0;
	for (f=0; f<FEATURES_MAX; f++)
	    for (v=0; v<VALUES_MAX; v++)
		count[c][f][v] = 0;
      }


    for (i=0; i<instances; i++)
    {
	data[i].trained = FALSE;
	data[i].weighted = FALSE;
	data[i].used = 1;
	data[i].correct = 1;
    }
}




/* ------------------------------------------------------------ */
/* PRINT_DISTANCE_TABLES (Prints the contents of all distance   */
/*   tables (For Debugging Purposes Only)			*/

void print_distance_tables(void)
{

    int f, v1, v2;
    int features, values;

    features = CONFIG.features;


    for (f=0; f<features; f++)
    {
	values = CONFIG.nvalues[f];
	printf("FEATURE: %d\n", f);
	for (v1=0; v1<values; v1++)
	{
	    for (v2=0; v2<values; v2++)
		printf("%.3f\t",dtable[f][v1][v2]);
	    printf("\n");
	}
	printf("\n\n");
    }
}


/* ------------------------------------------------------------ */

void print_count_table(void)
{
    int c,f,v;
    int classes,features,values;

    classes = CONFIG.classes;
    features = CONFIG.features;
    values = CONFIG.cont[0].discretes;

    printf("CONFIG: %d %d %d\n", classes, features, values);
    for (c=0; c<classes; c++)
    {
	printf("CLASS %d\n\n",c);
	for (v=0; v<values; v++)
	{
	    for (f=0; f<features; f++)
	        printf("%3d ", count[c][f][v]);
	    printf("\n");
        }
    }
}


/* ------------------------------------------------------------ */

void print_cprob_table(void)
{
    int c,f,v;
    int classes,features,values;

    classes = CONFIG.classes;
    features = CONFIG.features;
    values = CONFIG.cont[0].discretes;

    printf("CONFIG: %d %d %d\n", classes, features, values);
    for (c=0; c<classes; c++)
    {
	printf("CLASS %d\n\n",c);
	for (v=0; v<values; v++)
	{
	    for (f=0; f<features; f++)
	        printf("%5.2f ", cprob[c][f][v]);
	    printf("\n");
        }
    }
}




/* ------------------------------------------------------------ */
/* DTABLE_ENTRY: Computes distance table entry for a		*/
/*	specified feature and specified values			*/
/* INPUTS:  f  = feature #					*/
/*          v1 = value 1					*/
/*          v2 = value 2					*/


float dtable_entry(int f, int v1, int v2)
{
    int c, classes;
    float C1, C1i, C2, C2i, diff;
    float vdm = 0.0;
    float K;


    classes = CONFIG.classes;
    K = CONFIG.K;

    C1 = (float) count[classes][f][v1];
    C2 = (float) count[classes][f][v2];
    
    for (c=0; c<classes; c++)
    {
	C1i = (float) count[c][f][v1];
	C2i = (float) count[c][f][v2];

	if ((C1 == 0) || (C2 == 0)) 
	    diff = 1.0;
	else diff = (float) fabs(C1i/C1 - C2i/C2);

	if (K == 1) vdm += diff;
	else vdm += (float) pow(diff, K); 
    }

    return (vdm);
}



/* ------------------------------------------------------------ */
/* OVERLAP:  Overlap Distance Metric				*/
/*   Counts the number of feature value mismatches between      */
/*   two instances.						*/

float overlap(int i, int j)
{
    int mismatches;
    int f;
    int features = CONFIG.features;

    mismatches = 0;
    for (f=0; f<features; f++)
	if (data[i].value[f] != data[j].value[f])
	    mismatches++;

    return ((float) mismatches);
}


/* ------------------------------------------------------------ */
/* EUCLIDEAN:  Euclidean Distance Metric			*/

float euclidean(int i, int j)
{
    int f;
    int features = CONFIG.features;
    float distance = 0.0;
    float diff;

    for (f=0; f<features; f++)
    {
	diff = fabs(data[i].value[f] - data[j].value[f]);
        distance += diff*diff;
    }

    return (sqrt(distance));
}




/* ------------------------------------------------------------ */
/* MANHATTAN:  ManhattanDistance Metric				*/

float manhattan(int i, int j)
{
    int f;
    int features = CONFIG.features;
    float distance = 0.0;

    for (f=0; f<features; f++)
	distance += fabs(data[i].value[f] - data[j].value[f]);

    return (distance);
}




/* ------------------------------------------------------------ */
/* MVDM:  Modified Value Distance Metric			*/
/* INPUTS:  x = instance x					*/
/* 	    y = instance y					*/
/*          weighting = flag if weighting is ON/OFF		*/

float MVDM(int x, int y, int weighting)
{
    int i, xi, yi;
    float Wx, Wy, sum, R;
    int features;

    features = CONFIG.features;
    R = CONFIG.R;

    sum = 0.0;

    if (CONFIG.feature_weighting == OFF)
    {
        for (i=0; i<features; i++)
        {
 	  xi = data[x].value[i];
	  yi = data[y].value[i];
	  if (R==1) 
	   sum = sum + dtable[i][xi][yi];
	  else 
           sum = sum + (float) pow(dtable[i][xi][yi], R);
	}	
    }
    else
    {
        for (i=0; i<features; i++)
        {
 	  xi = data[x].value[i];
	  yi = data[y].value[i];
	  if (R==1) 
	   sum=sum+CONFIG.feature_weights[i] * dtable[i][xi][yi];
	  else 
           sum=sum+CONFIG.feature_weights[i]*(float) pow(dtable[i][xi][yi], R);
	}
    }

   if (weighting != OFF) 
   {
	Wx = data[x].weight;
	Wy = data[y].weight;
	if (Wy == 1.0) return (Wx * sum);
	else return (Wx * Wy * sum);
    }
   else return (sum);
}
    

/* ------------------------------------------------------------ */
/* BUILD_DISTANCE_TABLES:  Construct distance tables based on	*/
/*   statistical information accumulated during feature value   */
/*   counting.							*/

void build_distance_tables(void)
{
    int f,v1,v2;
    int values;
    int features = CONFIG.features;

    features = CONFIG.features;

    for (f=0; f<features; f++)
    {
	values = CONFIG.nvalues[f];
	for (v1=0; v1<values; v1++)
	    for (v2=0; v2<values; v2++)
	    {
	    	if (v1 == v2)  dtable[f][v1][v2] = 0.0;
    	    	else if (v1 > v2) dtable[f][v1][v2] = dtable[f][v2][v1];
	    	else dtable[f][v1][v2] = dtable_entry(f,v1,v2);
	    }
    }

}



/* ------------------------------------------------------------ */
/* BUILD_BAYES_TABLES:  Build table of conditional probabilities */

void build_bayes_tables(void)
{
    int c,f,v;
    int classes, features;

    int totval, totclass;

    classes = CONFIG.classes;
    features = CONFIG.features;

    totclass = 0;			/* Compute Prior Probabilities */
    for (c=0; c<classes; c++)
      totclass += class_count[c];
    
    for (c=0; c<classes; c++)
      prior[c] = (float) class_count[c] / (float) totclass;


					/* Compute Conditional Prob. */

    for (c=0; c<classes; c++)
        for (f=0; f<features; f++)
	{
	    totval = 0;
	    for (v=0; v<VALUES_MAX; v++)
		totval += count[c][f][v];

	    for (v=0; v<VALUES_MAX; v++)
		if (totval == 0) cprob[c][f][v] = prior[c] / (float) totclass;
		else cprob[c][f][v] = (float) count[c][f][v] / (float) totval;

	}
}



/* ------------------------------------------------------------ */
/* BAYESIAN CLASSIFIER:  Returns the probability of each class	*/
/*    PROBABILITIES ARE NOT CURRENTLY NORMALIZED                */

void compute_bayesian_probability(int instance, float classprob[])
{

    int i,j,k;
    int classes;
    float tprob, sumprob;
    int value, features;
   
    

    classes = CONFIG.classes;
    features = CONFIG.features;

    for (i=0; i<CLASSES_MAX; i++)
	classprob[i] = 0.0;

    sumprob = 0.0;
    for (k=0; k<classes; k++)
    {
        tprob = 1.0;
	for (j=0; j<features; j++)
	{
	    value = data[instance].value[j];
	    tprob *= cprob[k][j][value];
	}
	classprob[k] = tprob * prior[k];
	sumprob += classprob[k];
    }

}

