/* Copyright (c) 1993 by The Johns Hopkins University */


 
/* 

 PEBLS:  Parallel Exemplar-Based Learning System

 For more information, contact:				

 Steven Salzberg  (salzberg@cs.jhu.edu)
 Dept. of Computer Science
 Johns Hopkins University
 Baltimore, MD  21210


 Code written by:

 John N. Rachlin
 Dept. of Computer Science
 Johns Hopkins University
 Baltimore, MD  21210

*/



/* PEBLS.C:  Core routines for invoking various system modules */
/*           Provides routines for training and testing.       */



#include <stdio.h>
#include "config.h"
#include "pebls.h"



config_type   	CONFIG;   			     
instance_type 	data[INSTANCES_MAX];		     
int		class_count[CLASSES_MAX];
float		prior[CLASSES_MAX];
int	 	count[CLASSES_MAX+1][FEATURES_MAX][VALUES_MAX];
float	 	dtable[FEATURES_MAX][VALUES_MAX][VALUES_MAX];  
float		cprob[CLASSES_MAX][FEATURES_MAX][VALUES_MAX];
output_type 	output[CLASSES_MAX+1][TRIALS_MAX];   



/* ------------------------------------------------------------ */
/* TRAIN_INSTANCE:  Add instance "i" to exemplar data		*/
/* INPUTS:	i = current instance				*/
/* OUTPUT:	None.  Global COUNT data updated		*/

void train_instance(int i)
{
    int c, f, v;
    int classes = CONFIG.classes;
    int features = CONFIG.features;

    c = data[i].class_true;
    for (f=0; f<features; f++)		/* Update value counts */
    {
	v = data[i].value[f];
	count[c][f][v]++;
	count[classes][f][v]++;
    }

    class_count[c]++;

    data[i].trained = TRUE;
}







/* ------------------------------------------------------------ */
/*  LEAVE_ONE_OUT:  A training and test method by which    	*/
/*   each instance is tested by first training on all other     */
/*   instances in the data set.					*/
/*  INPUTS:  None.						*/
/*  OUTPUT:  None.  (Results printed to screen.)		*/

void leave_one_out(void)
{
    int i,j,t;
    int trials = CONFIG.trials;
    int instances = CONFIG.instances;
    int classes = CONFIG.classes;
    int weighting = CONFIG.exemplar_weighting;
    int nearest_class, nearest_list[K_NEIGHBOR_MAX];
    int k = CONFIG.nearest_neighbor;
    float distances[K_NEIGHBOR_MAX];
    float class_probability[CLASSES_MAX];

    if (CONFIG.output_mode == COMPLETE)
    {
	printf("\n\n%10s %10s %10s %10s\n", "", "", "TRUE", "PEBLS");
	printf("%10s %10s %10s %10s\n", "TRIAL", "ID", "CLASS", "CLASS");
	printf("%10s %10s %10s %10s\n", "-----", "-----", "-----", "-----");
    }

    for (i=0; i<instances; i++)  		/* For ea. instance */
    {
	initialize_training();
	for (j=0; j<instances; j++)		/* Train on all others */
	    if (j != i) train_instance(j);


	if (CONFIG.classify_mode == PEBLS) 
	    build_distance_tables();
	else if (CONFIG.classify_mode == BAYES)
	    build_bayes_tables();


        for (t=0; t<trials; t++)
        {

	    if (CONFIG.classify_mode == BAYES)
	    {
	        compute_bayesian_probability(i, class_probability);
		nearest_class = fmaximum(class_probability, classes);
	    }

	    else 

	    {
	        set_exemplar_weights(); 
	        nearest_neighbor(i, k, nearest_list, distances, weighting);
	        nearest_class = nearest_vote(k, nearest_list, distances);
	    }

            update_single_output(i, nearest_class, t);
	    if (CONFIG.output_mode == COMPLETE)
    	        printf("%10d %10s %10s %10s \n", 
			t+1,
		       	data[i].id,
		       	CONFIG.class_name[data[i].class_true], 
		       	CONFIG.class_name[nearest_class]);
	}
    }

    print_output();
}



/* ------------------------------------------------------------ */
/* TRAIN_SUBSET:  Train on a random subset of the instance data */
/* INPUT:  None.						*/
/* OUTPUT: None.						*/


void train_subset(void)
{
    int i, size;
    int select, select_order[INSTANCES_MAX];
    int training_instances;
    float training_size = CONFIG.training_size;

    if (CONFIG.training_mode == SUBSET)
	training_instances = CONFIG.instances;  	 /* Choose among ALL instances */
    else training_instances = CONFIG.training_instances; /* Choose among specified group */

    shuffle(select_order, training_instances);
    size = round (training_instances * training_size);  /* Actual number to train */

    for (i=0; i<size; i++)
    {
	select = select_order[i];
	train_instance(select);
    }
}




/* ------------------------------------------------------------ */
/* TRAIN_SPECIFIED_GROUP:  Train a specified group of instances */
/* INPUTS: None.						*/
/* OUTPUT: None.  						*/

void train_specified_group(void)
{
    int i, training_instances = CONFIG.training_instances;
    float training_size = CONFIG.training_size;

    if (training_instances == 0)
	error(NO_TRAIN_ERR, NULL);
    
    if (training_size == 1.00)
    {
        for (i=0; i<training_instances; i++)
	    train_instance(i);
    }
    else train_subset();
}




/* ------------------------------------------------------------ */
/* TRAIN:  Train instances. Invoke the proper training routines */
/* INPUTS: None.						*/
/* OUTPUT: None. (Global distance tables are produced		*/

void train(void)
{
    initialize_training();

    switch (CONFIG.training_mode)
    {
        case SUBSET: 
          train_subset(); 
	  break;

	case SPECIFIED_GROUP: 
	  train_specified_group(); 
	  break;
    }

    if (CONFIG.classify_mode == PEBLS) build_distance_tables();
    else if (CONFIG.classify_mode == BAYES) build_bayes_tables();
}




/* ------------------------------------------------------------ */
/* TEST:  Test all untrained instances				*/
/* INPUTS:  trial = # of the current trial;			*/

void test(void)
{
    int i;
    int instances = CONFIG.instances;
    int classes = CONFIG.classes;
    int k = CONFIG.nearest_neighbor;
    int nearest_list[K_NEIGHBOR_MAX];
    float distances[K_NEIGHBOR_MAX];
    float class_probability[CLASSES_MAX];


    for (i=0; i<instances; i++)
    {
       	if (data[i].trained == FALSE) 
	{
	    if (CONFIG.classify_mode == BAYES)
	    {
		compute_bayesian_probability(i, class_probability);
		data[i].class_nearest = fmaximum(class_probability, classes);
	    }
	    else
	    {
 	        nearest_neighbor(i, k, nearest_list, distances, 
				 CONFIG.exemplar_weighting);
	        data[i].class_nearest = nearest_vote(k, nearest_list, distances);
	    }
	}
    }

    post_process();
}








/* ============================================================ */

main(int argc, char *argv[])
{

    int t, trials, train_once=FALSE, weight_once=FALSE;

    if (argc != 2)  error(USAGE_ERR, NULL);
    else 
    {
      	initialize(argv[1]);
	if (CONFIG.training_mode == LEAVE_ONE_OUT)
	    leave_one_out();
	else
	{
	    trials = CONFIG.trials;

				/* IF TRAINING SET NEVER CHANGES, TRAIN ONCE */

	    if ((CONFIG.training_mode == SPECIFIED_GROUP) &&
		(CONFIG.training_size == 1.00)) 
	    {
	        train();
		train_once = TRUE;
	    }
				/* IF EXEMPLAR WEIGHTS NEVER CHANGE, */
				/* SET WEIGHTS ONCE		     */

	    if ((CONFIG.exemplar_weighting == ONE_PASS) &&
	        (CONFIG.classify_mode == PEBLS))
	    {
		set_exemplar_weights();
		weight_once = TRUE;
	    }


	    for (t=0; t<trials; t++)
	    {
		if (train_once == FALSE) train();
		if ((weight_once == FALSE) && (CONFIG.classify_mode == PEBLS)) 
		    set_exemplar_weights();
	        test();
		update_output(t);
	    }

	    print_output();
	}
    }
}


